package PureMessage::RPC::Client;
use PureMessage::RPC::Result;
use PureMessage::ServerGroup; # to get a %sg
use IO::Select;
use Data::Dumper;
use HTTP::Request::Common qw( POST $DYNAMIC_FILE_UPLOAD );
use strict;

$DYNAMIC_FILE_UPLOAD = 1;


=item new(\%hosts)

The paramater %hosts is a hash where the keys are the host names and the values are 
the information of the same as format as would be returned by 
C<PureMessage::ServerGroup::get_host>


        my ($h, $data) = PureMessage::ServerGroup::get_host($host);
        my $b = PureMessage::RPC::Client->new({$h => $data});



=cut

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my %self = ();
    $self{hosts} = shift;
    return unless ref($self{hosts}) eq "HASH";
    my %opts = @_;
    $self{timeout} = $opts{timeout} || $PureMessage::ServerGroup::sg{rpc_timeout} || 5;
    $self{socket} = {};
    $self{select} = {};
    $self{request_type} = "GET";
    $self{headers} = {};
    return bless( \%self, $class);
}

sub event {
    # XXX use this to register custom callbacks
}

sub call {
    my $self = shift;
    my ($service, @args) = @_;
    
    $service =~ s/\./\//;
    # XXX raise_error if ! $service
    $self->{url} = "/RPC/$service/" . join("/", map { urlize($_) } @args);
    $self->{error} = {};
    $self->{select} = IO::Select->new();
    while ( my ($host, $info) = each %{$self->{hosts}} ) {
        $self->{result}{$host} = PureMessage::RPC::Result->new();
        unless ( $info->{proto} ) {
            require PureMessage::ServerGroup;
            $info->{proto} = PureMessage::ServerGroup::proto($info);
        }
        my $socket;
        if ( $info->{proto} eq "HTTPS" ) {
            require Net::HTTPS::NB;
            $socket = Net::HTTPS::NB->new( Host => $info->{host} . ":" . $info->{port}, 
                                           SSL_verify_mode => 0 ) 
                                || ($self->raise_error($host, $!) and next);
        }
        elsif ( $info->{proto} eq "HTTP" ) {
            require Net::HTTP::NB;
            $socket = Net::HTTP::NB->new( Host => $info->{host} . ":" . $info->{port} )
                                || ($self->raise_error($host, $!) and next);
        }
        else {
            $self->{error}{$host} = "Can't connect $!";
            next;
        }
	if ( $socket ) {
            $self->{socket}{$socket} = $host;
            if ( $self->{request_type} eq "GET" ) {
                $socket->write_request($self->{request_type}, 
                        $self->{url},
                        %{$self->{headers}},
                        Cookie => $info->{auth} ? "sid=$info->{auth}, path=/" : "");
            }
            elsif ( $self->{request_type} eq "POST" ) {
                my $req = POST $self->{url},
                          Content_Type => 'form-data',
                          Content      => $self->{attachments};
                my $cont_ref = $req->content;
                $socket->write_request($self->{request_type}, 
                        $self->{url},
                        %{$self->{headers}},
                        %{$req->headers()},
                        Cookie => $info->{auth} ? "sid=$info->{auth}, path=/" : "");
                my $buf;
                while( ($buf = &$cont_ref()), defined($buf) && length($buf)) {
                    $socket->print($buf);
                }
            }
            else {
                die "Unsupported request type: $self->{request_type}";
            }
            $self->{select}->add($socket);
        }
        else {
            $self->{error}{$host} = "Can't connect $!";
        }
    }
}

sub wait {
    my $self = shift;
    while ( my @ready = $self->{select}->can_read($self->{timeout}) ) {
        foreach my $socket ( @ready ) {
            my $host = $self->{socket}{$socket};
            next if $self->{result}{$host}{eof};
            if ( ! $self->{result}{$host}{code} ) {
                # expect headers
                my ($code, $mess, %h);
                eval {
                    ($code, $mess, %h) = $socket->read_response_headers();
                };
                if ( $@ ) {
                    $code = "501";
                    $mess = $!;
                    %h = ('X-Perl-Error' => $@);
                }
                next unless defined $code;
                $self->{result}{$host}{code} = $code;
                $self->{result}{$host}{msg} = $mess;
                $self->{result}{$host}{headers} = \%h;
                if ( $code != 200 ) {
                    if ( $code == 403 ) {
                        $self->raise_error($host, "Access denied");
                    }
                    elsif ( $code == 401 ) {
                        $self->raise_error($host, "Authentication failed");
                    }
                    else {
                        $self->raise_error($host, "Unknown error $code");
                    }
                }
            }
            else {
                my $buf;
                my $n = $socket->read_entity_body($buf, 1024);
                $self->{error}{$host} = "Body read error" if ! defined $n;
                $self->{result}{$host}{eof} = 1 if defined $n and $n == 0;
                $self->{result}{$host}->parse($buf, 1);
                # XXX do data callback here
            }
        }
        my $status = 1;
        foreach my $host ( keys %{$self->{hosts}} ) {
            $status = 0 if not ( $self->{error}{$host} or $self->{result}{$host}{eof} );
        }
        last if $status;
    }
}

sub attach {
    # this is needed for Publishing API
    my $self = shift;
    push @{$self->{attachments}}, @_;
    $self->{request_type} = "POST";
}

sub data {
    my $self = shift;
    my $host = shift;
    if ( $host ) {
        return $self->{result}{$host}->data() if $self->{result}{$host};
    }
    else {
        return $self->{result};
    }
}

sub error {
    my $self = shift;
    my $host = shift;
    my @hosts = $host ? ($host) : keys %{$self->{hosts}};
    my @errors;
    foreach ( @hosts ) {
        my $err;
        if ( $err = $self->{error}{$_} ) {
            push @errors, scalar @hosts == 1 ? $err : "$_: " . $err;
        }
        elsif ( $self->{result}{$_} and $err = $self->{result}{$_}->error() ) {
            push @errors, scalar @hosts == 1 ? $err : "$_: " . $err;
        }
    }
    return 0 if scalar @errors < 1;
    return join("\n", @errors);
}

sub raise_error {
    my $self = shift;
    my $host = shift;
    $self->{error}{$host} = join(" ", @_);
}

sub urlize {
  my $rv = shift;
  $rv =~ s/([^A-Za-z0-9])/sprintf("%%%2.2X", ord($1))/ge;
  return $rv;
}

1;
