package PureMessage::ServerGroup;
use Carp;
use strict;
use PureMessage::Version qw($PREFIX);
use ActiveState::Config qw(:path);
use PureMessage::_sys;
use PureMessage::Manager::HostDiscovery qw( &discovery_probe_reply_decode
                                            &discovery_probe_old_reply_decode
                                          );
use PureMessage::ServerGroup::Utils qw(to_ipaddress);
use PureMessage::ServerGroup::Discover qw(scan_network determine_proto);

use Socket;
use base 'Exporter';
use vars qw(@EXPORT @EXPORT_OK %EXPORT_TAGS);
our @EXPORT_OK = qw(read_config save_group get_groups save_host get_hosts 
                    del_host membership del_group rename_host rename_group
                    get_host get_group proto network_scan is_self $error);
%EXPORT_TAGS = (all => \@EXPORT_OK);

our %sg;
our $error = "";
my $p;
my $conf;
my $file = "$PREFIX/etc/servergroups.conf";
my $writable = 0;
read_config() if -e $file;

sub read_config {
    my %opts = @_;
    $file = $opts{file} || $file;
    $p = ActiveState::Config->new;
    $writable = $opts{writable} || 0;
    $p->typedecl("group#*.hosts", "#list");
    $p->typedecl("rpc_timeout", "#duration");
    $p->default("rpc_timeout",'5s');
    eval {
        $conf = $p->parse_file($file, writable => $writable);
    };
    if ( $@ ) {
        $error = "Can't open file $file for writing.";
    }
    %sg = $p->as_hash();
    return \%sg;
}

sub save_group {
    my $group = shift;
    my $data = shift;
    croak "You must specify a group name"
        unless($group);
    read_config(writable=>1) unless $writable;
    my $node = _add_section($p->root, "group", $group);
    while ( my ($k, $v) = each (%{$data}) ) {
        my $opt = path_node($p->root, "group#$group.$k", create => 'value');
        $opt->value($v);
        $node->attach($k, $opt);
    }
    path_node($p->root, "group#$group.hosts", create => 'value');
    $p->typedecl("group#$group.hosts", "#list");
    save();
}

sub save_host {
    my $host = shift;
    my $data = shift;
    croak "You must specify a host alias"
        unless($host);
    croak "You must specify a host name"
        unless($data->{host});
    read_config(writable=>1) unless $writable;
    my $node = _add_section($p->root, "host", $host);
    if ( $data->{user} ) {
        # get auth
        my $auth = get_auth_key($data);
        if ( ! $auth ) {
            return 0;
        }
        $data->{auth} = $auth;
    }
    foreach my $k ( 'description', 'host', 'port', 'auth', 'user', 'proto' ) {
        next unless $data->{$k};
        my $opt = path_node($p->root, "host#$host.$k", create => 'value');
        $opt->value($data->{$k});
        $node->attach($k, $opt);
    }

    # PM 4.5 introduces an alternate rpc/syncronization system that uses a
    # different process than the Mananger RPC. By default this runs on TCP port
    # 28080 and uses HTTP.
    my $sync_port = $data->{sync_port} 
		  || $sg{host}{$host}{sync_port} 
		  || 28080;
    my $sync_proto = $data->{sync_proto} 
		  || $sg{host}{$host}{sync_proto}
		  || "HTTP";
    path_value($p->root, "host#$host.sync_port", $sync_port);
    path_value($p->root, "host#$host.sync_proto", $sync_proto);
    
    save();
    if ( $data->{group} ) {
        my %grps = map { $_ => 1 } ref($data->{group}) eq "ARRAY" ? @{$data->{group}} : ($data->{group});
        membership($host, \%grps);
    }
    return 1;
}

sub rename_group {
    my $group = shift;
    my $new_name = shift;
    read_config(writable=>1) unless $writable;
    my $node = path_node($p->root, "group#$group");
    if ( ! $node ) {
        raise_error("Unknown group: $group");
        return;
    }
    $node->name($new_name);
    $node->path("group#$new_name");
    save();
}

sub rename_host {
    my $host = shift;
    my $new_name = shift;
    
    read_config(writable=>1) unless $writable;
    my $node = path_node($p->root, "host#$host");
    if ( ! $node ) {
        raise_error("Unknown host: $host");
        return;
    }
    $node->name($new_name);
    $node->path("host#$new_name");
    my $groups = membership($host);
    membership($new_name, $groups); # add new one
    membership($host, {}); # remove old name from groups
    save();
}

sub get_auth_key {
    my $data = shift;
    unless ( $data->{proto} ) {
        $data->{proto} = proto($data);
        if ( ! $data->{proto} ) {
            return undef;
        }
    }
    my $host = $data->{'host'};
    $host .= ":" . $data->{'port'} if $data->{'port'};
    my $s;
    if ( $data->{proto} eq "HTTPS" ) {
        require Net::HTTPS;
        $s = Net::HTTPS->new( Host => $host, SSL_verify_mode=>0) 
                            || (raise_error($!) and return);
    }
    elsif ( $data->{proto} eq "HTTP" ) {
        require Net::HTTP;
        $s = Net::HTTP->new( Host => $host )
                            || (raise_error($!) and return);
    }
    else {
        raise_error("Unsupported protocol $data->{proto}");
        return undef;
    }
    my $content = "user=$data->{user}&pass=$data->{pass}";
    $s->write_request(POST => '/session_login.cgi', 
                      "Cookie" => "testing=1;", 
                      "Content-Type" => "application/x-www-form-urlencoded", 
                      $content );
    my($code, $mess, %h) = $s->read_response_headers;
    
    if ( $h{'Set-Cookie'} ) {
        $h{'Set-Cookie'} =~ /^sid=(\w+);/;
        return $1 if $1;
    }
    raise_error("Authentication failed");
    return undef;

}

sub get_groups {
    my @unsorted_hosts = get_orphans();
    if ( scalar @unsorted_hosts > 0 ) {
        $sg{group}{"\0"} = { description => "Unsorted hosts", hosts=>\@unsorted_hosts };
    }
    return $sg{group};
}

sub get_hosts {
    my $group = shift;
    if ( $group and ref($group) eq "ARRAY" ) {
        my %hosts = map { $_ => $sg{host}{$_} } @{$group};
        return \%hosts;
    }
    elsif ($group) {
        return undef unless ref($sg{group}{$group}{hosts}) eq "ARRAY";
        my %hosts = map { $_ => $sg{host}{$_} } @{$sg{group}{$group}{hosts}};
        return \%hosts;
    }
    return $sg{host};
}

sub is_self {
    my $host = shift;
    my $my_host = shift || PureMessage::_sys::gethostname();
    my $port = shift || $ENV{SERVER_PORT} || 18080;
    my $host_ip = to_ipaddress($host->{host});
    my $my_ip = to_ipaddress($my_host);
    return 0 unless $host_ip eq $my_ip;
    return 0 unless $host->{port} eq $port;
    return 1;
}

sub get_orphans {
    return  unless $sg{host};
    my %hosts = %{$sg{host}};
    foreach ( keys %{$sg{group}} ) {
        next unless ref($sg{group}{$_}{hosts}) eq "ARRAY";
        foreach my $h ( @{$sg{group}{$_}{hosts}} ) {
            delete $hosts{$h};
        }
    }
    return keys %hosts;
}

sub get_host {
    my $h = shift;
    read_config() unless $sg{host} and $writable;
    if ( ref ($h) eq "HASH" ) {
        while ( my ($id, $host) = each (%{$sg{host}}) ) {
            if ( (not defined $h->{port} or $host->{port} == $h->{port}) 
                  and lc($host->{host}) eq lc($h->{host}) ) {
                return wantarray ? ($id=>$host) : $host;
            }
        }
        return undef;
    }
    else {
        return undef unless $sg{host}{$h};
        return wantarray ? ($h=>$sg{host}{$h}) : $sg{host}{$h};
    }
}

sub get_group {
    my $g = shift;
    read_config() unless $sg{group} and $writable;
    return $sg{group}{$g} || undef;
}

sub del_group {
    my $group = shift;
    # delete section
    read_config(writable=>1) unless $writable;
    my $iter= $p->root->children;
    while (my $node = $iter->next) {
        next unless $node->groupclass eq 'group';
        next unless $node->name eq "$group";
        $iter->delete();
        last;
    }
    save();
}

sub del_host {
    my $host = shift;
    read_config(writable=>1) unless $writable;
    # delete section
    my $iter= $p->root->children;
    while (my $node = $iter->next) {
        next unless $node->groupclass eq 'host';
        next unless $node->name eq "$host";
        $iter->delete();
        last;
    }
    # clear memberships
    membership($host, {});
    save();
}

sub membership {
    my $host = shift;
    my $set_grps = shift;
    read_config(writable=>1) if $set_grps and not $writable;
    my %ret_grps;
    foreach my $group ( keys %{$sg{group}} ) {
        if ( my $m = path_node($p->root, "group#$group") ) {
            my $h = path_node($p->root, "group#$group.hosts", create => 'value');
            my $hosts = path_value($p->root, "group#$group.hosts");
            my %hosts = map { $_ => 1 } @$hosts;
            if ( $set_grps ) {
                if ( $set_grps->{$group} ) {
                    $hosts{$host} = 1;
                }
                else {
                    delete $hosts{$host};
                }
                my @new_hosts = keys %hosts;
                path_value($p->root, "group#$group.hosts", \@new_hosts);
            }
            else {
                $ret_grps{$group} = 0;
                foreach ( @{$hosts} ) { 
                    $ret_grps{$group} = 1 if $_ eq $host;
                }
            }
        }
    }
    if ( $set_grps ) {
        save();
        return $set_grps;
    }
    return \%ret_grps;
}

sub proto {
    my %opts = (host => shift);

    # look for a new style discovery probe, if not found try and
    # decode an old style response
    $opts{decoders} = [ \&discovery_probe_reply_decode,
                        \&discovery_probe_old_reply_decode
                      ];

    my $ret;
    eval {
        $ret = determine_proto(%opts);
    };
    if ($@) {
        $error = $@;
    }
    return $ret;
}

sub _add_section {
    my $parent = shift;
    my $class = shift;
    my $name = shift;
    my $path = $class;
    $path .= "#$name" if $name;
    $path = $parent->path() . ".$path";
    my $node = path_node($p->root, $path, create => 'group');
    die "Non object $path" unless (defined($node) && ref($node) =~ /^ActiveState::Config/);
    $node->groupclass($class);
    $node->name($name);
    $node->path($path);
    return $node;
}

# scan the network for webmin servers
# for each one found, add it to the cache.
sub network_scan {
    my %opts = @_;

    # setup save callback
    if ($opts{save}) {
        $opts{save} = sub { 
            my ($host, $hostref) = @_;
            save_host($host, $hostref) unless get_host($hostref);
        };
    }

    $opts{decoders} = [ \&discovery_probe_reply_decode,
                        \&discovery_probe_old_reply_decode
                      ];

    my $rc;
    eval {
        $rc = scan_network(%opts);
    };
    if ($@) {
        $error = $@;
    }
    return $rc;
}

sub save {
    my $filename = shift;
    if ( $filename ) {
        $p->saveto($filename);
    }
    else {
        $p->save();
    }
    read_config();
}

sub raise_error {
    $error = shift;
    return;
}

1;
