package PureMessage::ServerGroup::Discover;
use Carp;
use strict;

use PureMessage::ServerGroup::Utils;
use Socket;
use base 'Exporter';
use vars qw(@EXPORT_OK);
our @EXPORT_OK = qw(determine_proto scan_network);

# scan the network for webmin servers
# for each one found, add it to the cache.
sub scan_network {
    my %opts = @_;

    die "decoders is not set" unless defined $opts{decoders};
    my $port = $opts{port} || 18080;
    socket(BROAD, PF_INET, SOCK_DGRAM, getprotobyname("udp")) || 
        die("Socket failed",$!) && return;
    setsockopt(BROAD, SOL_SOCKET, SO_BROADCAST, pack("l", 1));
    my $myhostname = PureMessage::_sys::gethostname();
    my $myip = to_ipaddress($myhostname);
    my @broad;
    my %myaddr;
    if ($myip =~ /^(\d+\.\d+\.\d+)\.\d+$/) {
            if($1 ne "127.0.0"){ # don't add the loopback address
                    push(@broad, "$1.255");
                    $myaddr{inet_aton($myip)}++;
            }
    }
    push(@broad, "255.255.255.255");
    my $limit = $opts{time_limit} || 8;
    # send out the packets
    my %sent;
    foreach $b ( @broad ) {
        next if $sent{$b};
        send(BROAD, "webmin", 0, sockaddr_in($port, inet_aton($b)));
        $sent{$b}++;
    }
    # Get responses
    $| = 1;
    my $tmstart = time();
    my %servers;
    my %already;
    while (time()-$tmstart < $limit) {
        my $rin = "";
        vec($rin, fileno(BROAD), 1) = 1;
        if (select($rin, undef, undef, 1)) {
            my $buf;
            my $from = recv(BROAD, $buf, 1024, 0);
            my %host_data;
            foreach (@{$opts{decoders}}) {
                %host_data = $_->($buf);
                last if %host_data;
            }

            next if (!$from);
            my ($fromport, $fromaddr) = sockaddr_in($from);
            my $fromip = inet_ntoa($fromaddr);
            $fromport = $host_data{port} || $fromport; # use the port from reply if it's there

            if ($fromip !~ /\.(255|0)$/ && !$already{$fromip}++) {
                my $byname = gethostbyaddr($fromaddr, AF_INET);
                # if a reverse fails, use the IP address instead
                if(!$byname){
                    $byname = $fromip;
                }
                else {
                    # check to see if this is us ...
                    # FIXME: if we have other IP addresses bound (that responded) and there isn't any DNS information
                    # for them, this check will not work.
                    if( $byname =~ /$myhostname/i){
                        # We're talking to ourselves!
                        next unless $opts{localhost};
                    }
                }
                # add server
                my %host = ( 
                                host => lc($byname),
                                port => $fromport,
                                description => "PureMessage host $byname:$fromport",
                                proto => $host_data{use_ssl} ? 'HTTPS' : 'HTTP',
                                has_defaults => $host_data{has_defaults},
                            );
                # prepare host alias and make sure ther are no dots in there (e.g. if IP is used)
                $byname = $1 if $byname =~ /^(\w+)\./;
                $byname =~ s/\./_/g;
                $servers{"$byname:$port"} = \%host;
                if ( $opts{save} ) {
                    $opts{save}->("$byname:$port", \%host);
                }
            }
        }
    }
    return \%servers;
}

sub determine_proto {
    my %opts = @_;

    die "decoders is not set" unless defined $opts{decoders};
    my $host = $opts{host};
    unless ( ref($host) eq "HASH" ) {
        # XXX read port from config file (?)
        $host = { host => $host, port => '18080' };
    }
    return die "Bad connection profile." 
        unless $host->{host} and $host->{port};
    socket(S, PF_INET, SOCK_DGRAM, getprotobyname("udp")) 
        || return die "Can't create UDP socket: $!";
    my $ipaddr = inet_aton($host->{host});
    return die "Can't resolve $host->{host}" unless $ipaddr;
    my $portaddr = sockaddr_in($host->{port}, $ipaddr);
    send(S, 'webmin',0,$portaddr) == length('webmin') 
        || return die "Cannot send to udp $host->{host}:$host->{port}: $!";
    my $data = '';
    my $limit = 2;
    $| = 1;
    my $tmstart = time();
    while(time()-$tmstart < $limit) {
            my $rin = "";
            vec($rin, fileno(S), 1) = 1;
            if (select($rin, undef, undef, 1)) {
                    my $n = recv(S, $data, 100, 0) or 
                        return die "Failed to receive UDP data from $host->{host}";
                    last;
            }
    }
    close S;
    return die "Failed to get UDP response from $host->{host}." unless $data;

    my %host_data;
    foreach (@{$opts{decoders}}) {
        %host_data = $_->($data);
        last if %host_data;
    }
    return $host_data{use_ssl} ? "HTTPS" : "HTTP";
}

1;
